#############################################################################
# readdvd.tcl Copyright 2004 r0lZ
# read DVD IFOs in memory.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::readdvd {

namespace export readdvd readdvdfile get_int parse_PGCI_UT

#############################################################################
## Procedure:  get_int
## Get an integer from inside the binary data
proc {get_int} {binarystring_upvar start_offset length} {

	upvar $binarystring_upvar binarystring

	set s [string range $binarystring $start_offset [expr $start_offset + $length - 1]]
	binary scan $s H* hex
	return [expr 0x$hex]
}

#############################################################################
## Procedure:  get_hex_s
## Get a hex string from inside the binary data.  Pad bytes with spaces.
proc {get_hex_s} {binarystring_upvar start_offset length} {

	upvar $binarystring_upvar binarystring

	set s [string range $binarystring $start_offset [expr $start_offset + $length - 1]]
	binary scan $s H* hex
	set len [string length $hex]
	set out ""
	for {set i 0} {$i<$len} {incr i 2} {
		set out "$out[string range $hex $i [expr $i+1]] "
	}
	return [string trim $out]
}

#############################################################################
## Procedure:  getCommandsFromPGC
## return a list with all commands of the provided PGC in hexadecimal form.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
## Elements beginning with * are pre, post and cell separators
proc {getCommandsFromPGC} {data_upvar pgcindex} {

	upvar $data_upvar data

	set cmdtable_start [get_int data [expr $pgcindex+0xE4] 2]
	if {$cmdtable_start == 0} {
		set num_pre_commands  0
		set num_post_commands 0
		set num_cell_commands 0
	} else {
		# remove PGC header, and keep only the command table and whatever follow...
		set cmds_table [string range $data [expr $pgcindex+$cmdtable_start] end]

		# get number of pre, post and cell commands
		set num_pre_commands  [get_int cmds_table 0 2]
		set num_post_commands [get_int cmds_table 2 2]
		set num_cell_commands [get_int cmds_table 4 2]
	}

	set idx 8
	lappend outlist "* pre"
	for {set i 0} {$i < $num_pre_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	lappend outlist "* post"
	for {set i 0} {$i < $num_post_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	lappend outlist "* cell"
	for {set i 0} {$i < $num_cell_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	return $outlist
}

#############################################################################
## Procedure:  getHeaderFromPGC
## return the header part of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getHeaderFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	return [string range $data $pgcindex [expr $pgcindex + 0xEB]]
}

#############################################################################
## Procedure:  getProgramMapFromPGC
## return the program map part of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getProgramMapFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numprograms [get_int data [expr $pgcindex + 2] 1]
	if {$numprograms == 0} {return [list 0 ""]}
	set start [expr $pgcindex + [get_int data [expr $pgcindex + 0xE6] 2]]
	set end [expr $start + $numprograms - 1]
	if {[expr $end % 2] == 0} {incr end}	;# must be padded to word boundary
	return [list $numprograms [string range $data $start $end]]
}

#############################################################################
## Procedure:  getCellsPBTableFromPGC
## return the cells Playback table out of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getCellsPBTableFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numcells [get_int data [expr $pgcindex + 3] 1]
	if {$numcells == 0} {return [list 0 ""]}
	set start [expr $pgcindex + [get_int data [expr $pgcindex + 0xE8] 2]]
	set end [expr $start + ($numcells * 24) - 1]
	return [list $numcells [string range $data $start $end]]
}

#############################################################################
## Procedure:  getCellsPosTableFromPGC
## return the cells Position table out of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getCellsPosTableFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numcells [get_int data [expr $pgcindex + 3] 1]
	if {$numcells == 0} {return ""}
	set start [expr $pgcindex + [get_int data [expr $pgcindex + 0xEA] 2]]
	set end [expr $start + ($numcells * 4) - 1]
	return [string range $data $start $end]
}

#############################################################################
## Procedure:  parse_PGCI_UT
## process the data in the VMGM_PGCI_UT
proc {parse_PGCI_UT} {pgci_ut_raw_data_upvar vtsnum menusvar} {

	upvar $pgci_ut_raw_data_upvar pgci_ut_raw_data

	# get number of Language Units
	if {$pgci_ut_raw_data == ""} {
		set ${menusvar}($vtsnum,numlus) 0
		return
	} else {
		set numlus [get_int pgci_ut_raw_data 0 2]
		set ${menusvar}($vtsnum,numlus) $numlus
	}

	#dprint 2 "Number of Language Units: $numlus"


	# for all Language units
	for {set lu 1} {$lu <= $numlus} {incr lu} {

		# get LU's language, number of PGCs and offset
		set lang [string range $pgci_ut_raw_data [expr $lu * 8] [expr ($lu * 8) + 1]]
		if {$lang == "\000\000" || $lang == "\777\777" || $lang == "  "} {set lang "--"}
		set ${menusvar}($vtsnum,$lu,lang) $lang

		set offset [get_int pgci_ut_raw_data [expr ($lu * 8) + 4] 4]

		set ludata [string range $pgci_ut_raw_data $offset end]

		set numpgcs [get_int ludata 0 2]
		set ${menusvar}($vtsnum,$lu,numpgcs) $numpgcs

		#dprint 2 "\nMenu language unit $lu: \"$lang\""
		#dprint 2 "offset to VMGM_LU = 0x[format %08X $offset] ($offset)"


		# for each PGC
		for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {

			# get and store category mask
			set category [get_int ludata [expr $pgc * 8] 1]
			set ${menusvar}($vtsnum,$lu,$pgc,mask) $category
			switch [expr $category & 0x0F] {
				0 {set masktext ""}
				2 {set masktext "(Title)"}
				3 {set masktext "(Root)"}
				4 {set masktext "(SubPic)"}
				5 {set masktext "(Audio)"}
				6 {set masktext "(Angle)"}
				7 {set masktext "(Chapter)"}
				default {set masktext "(UNKNOWN)"}
			}
			set ${menusvar}($vtsnum,$lu,$pgc,masktext) $masktext

			# get and store PGC offset
			set pgc_offset [get_int ludata [expr ($pgc * 8) + 4] 4]

			# get and store PGC's commands
			set ${menusvar}($vtsnum,$lu,$pgc) [getCommandsFromPGC ludata $pgc_offset]

			# get and store the other PGC parts
			set ${menusvar}($vtsnum,$lu,$pgc,header) [getHeaderFromPGC ludata $pgc_offset]
			set tmp [getProgramMapFromPGC ludata $pgc_offset]
			set ${menusvar}($vtsnum,$lu,$pgc,numprograms) [lindex $tmp 0]
			set ${menusvar}($vtsnum,$lu,$pgc,programmap)  [lindex $tmp 1]

			set tmp [getCellsPBTableFromPGC ludata $pgc_offset]
			set ${menusvar}($vtsnum,$lu,$pgc,numcells)    [lindex $tmp 0]
			set ${menusvar}($vtsnum,$lu,$pgc,cellpbtable) [lindex $tmp 1]

			set ${menusvar}($vtsnum,$lu,$pgc,cellpostable) \
					[getCellsPosTableFromPGC ludata $pgc_offset]
		}
	}
}

#############################################################################
## Procedure:  parse_VTS_PGCI
## process the data in the VTS_PGCI of a VTS IFO
proc parse_VTS_PGCI {pgci_data_upvar vtsnum pgcsvar} {

	upvar $pgci_data_upvar pgci_data

	if {$pgci_data == ""} {
		set ${pgcsvar}($vtsnum,numpgcs) 0
		return
	}
	set numpgcs [get_int pgci_data 0 2]
	set ${pgcsvar}($vtsnum,numpgcs) $numpgcs

	# for each PGC
	for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
		# get and store PGC offset
		set pgc_offset [get_int pgci_data [expr ($pgc * 8) + 4] 4]
		# get and store PGC's commands
		set ${pgcsvar}($vtsnum,$pgc) [getCommandsFromPGC pgci_data $pgc_offset]

		set ${pgcsvar}($vtsnum,$pgc,header) [getHeaderFromPGC pgci_data $pgc_offset]

		set tmp [getProgramMapFromPGC pgci_data $pgc_offset]
		set ${pgcsvar}($vtsnum,$pgc,numprograms) [lindex $tmp 0]
		set ${pgcsvar}($vtsnum,$pgc,programmap)  [lindex $tmp 1]

		set tmp [getCellsPBTableFromPGC pgci_data $pgc_offset]
		set ${pgcsvar}($vtsnum,$pgc,numcells)    [lindex $tmp 0]
		set ${pgcsvar}($vtsnum,$pgc,cellpbtable) [lindex $tmp 1]

		set ${pgcsvar}($vtsnum,$pgc,cellpostable) [getCellsPosTableFromPGC pgci_data $pgc_offset]
	}
}

#############################################################################
## Procedure:  read_VMG_IFO
## read the data of the VIDEO_TS.IFO of the DVD directory supplied
proc read_VMG_IFO {filename sectorsvar pgcsvar menusvar} {
	#dprint 2 "Input Directory: $dvddir"

	set data [readFile $filename binary]
	if {$data == ""} {return 1}

	# Verify file is of correct type
	if {[string range $data 0 11] != "DVDVIDEO-VMG"} {
		errmsg 1000 "checking IFO file type" \
				"File \"$filename\"\nis not of correct type DVDVIDEO-VMG"
		return 1000
	}

	#dprint 2 ""
	##### split the VIDEO_TS.IFO file data in main tables parts
	set start_address(0) 0

	### WARNING: BUG in IfoEdit: the end address of the VMGM_MAT is not exact!!!
	### IfoEdit never change this address, even when the command table is growing!
	### We must therefore calculate the real end address (what a pitty!).
	set end_address(0) [get_int data 0x80 4]	;# should work, but...
	set fp_pgc_start [get_int data 0x84 4]
	set cmdtable_start [get_int data [expr $fp_pgc_start+0xE4] 2]
	set numpre  [get_int data [expr $fp_pgc_start+$cmdtable_start] 2]
	set numpost [get_int data [expr $fp_pgc_start+$cmdtable_start+2] 2]
	set numcell [get_int data [expr $fp_pgc_start+$cmdtable_start+4] 2]
	set cmdtable_len [expr 8+($numpre*8)+($numpost*8)+($numcell*8)]
	set cmdtable_end [expr $fp_pgc_start+$cmdtable_start+$cmdtable_len-1]
	if {$cmdtable_end != $end_address(0)} {
		set msg "Warning: The value for \"End byte of VMGI_MAT\" is incorrect!"
		set msg "$msg\n(This value is in the VIDEO_TS.IFO's VMGM_MAT at offset 0x80)"
		set msg "$msg\n\nValue in file: $end_address(0)"
		set msg "$msg\nCorrect value: $cmdtable_end"
		set msg "$msg\n\nIfoEdit doesn't upgrade this value when you add or remove"
		set msg "$msg\ncommands in the First-Play PGC."
		set msg "$msg\n\nThe correct value will be updated when you save the DVD."
		tk_messageBox -title "PgcEdit: Open DVD" -icon warning \
			-message $msg -parent . -type ok
		# finally, fix the error!
		if {$cmdtable_end > $end_address(0)} {
			set end_address(0) $cmdtable_end
		}
	}

	for {set i 1} {$i < 8} {incr i} {
		set start_address($i) [expr [get_int data [expr 0xC0 + (4 * $i)] 4] * 0x800]
		if {$start_address($i) == 0} {
			set end_address($i) 0
		} else {
			# VMGM_VOBU_ADMAP has the end_byte at position 0, and
			# VMG_TXTDT_MG at pos 16! All other tables at pos 4.
			switch $i {
				5 {set end_byte_pos 16}
				7 {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set end_address($i) [expr $start_address($i) + \
					[get_int data [expr $start_address($i)+$end_byte_pos] 4]]
		}
	}
	set i 0
	foreach p {VMGM_MAT VMG_PTT_SRPT VMGM_PGCI_UT VMG_PTL_MAIT \
			VMG_VTS_ATRT VMG_TXTDT_MG VMGM_C_ADT VMGM_VOBU_ADMAP} {
		if {$end_address($i) > 0} {
			set ${sectorsvar}(0,$p) [string range $data $start_address($i) \
						$end_address($i)]
		} else {
			set ${sectorsvar}(0,$p) ""
		}
		incr i
	}

	parse_VMG_IFO $sectorsvar $pgcsvar $menusvar

	return 0
}

#############################################################################
## Procedure:  parse_VMG_IFO
## process the data of the VIDEO_TS.IFO
proc parse_VMG_IFO {sectorsvar pgcsvar menusvar} {
	##### Number of title sets
	set num_vts [get_int ${sectorsvar}(0,VMGM_MAT) 0x3E 2]
	set ${pgcsvar}(numvts) $num_vts
	set ${menusvar}(numvts) $num_vts	;#!!!!!!


	##### VMGM_MAT/FP-PGC
	#dprint 2 "\nFirst Play PGC"
	set fp_pgc_start [get_int ${sectorsvar}(0,VMGM_MAT) 0x84 4]
	set ${pgcsvar}(0,numpgcs) 1
	set ${pgcsvar}(0,0) [getCommandsFromPGC ${sectorsvar}(0,VMGM_MAT) $fp_pgc_start]
	set ${pgcsvar}(0,0,header) [getHeaderFromPGC ${sectorsvar}(0,VMGM_MAT) $fp_pgc_start]
	set ${pgcsvar}(0,0,numprograms)  0
	set ${pgcsvar}(0,0,programmap)   ""
	set ${pgcsvar}(0,0,numcells)     0
	set ${pgcsvar}(0,0,cellpbtable)  ""
	set ${pgcsvar}(0,0,cellpostable) ""

	##### VMGM_PGCI_UT
	#dprint 2 "\nVMGM_PGCI_UT"
	parse_PGCI_UT ${sectorsvar}(0,VMGM_PGCI_UT) 0 ${menusvar}


	##### get the table of titles from the VMG_PTT_SRPT
	set data [subst $${sectorsvar}(0,VMG_PTT_SRPT)]
	if {$data != ""} {
		set numtitles [get_int data 0 2]
		for {set t 1} {$t <= $numtitles} {incr t} {
			set t_idx [expr 8+(($t-1)*12)]
			set ${pgcsvar}([get_int data [expr $t_idx+6] 1],[get_int data [expr $t_idx+7] 1],Title) $t
		}
	}
	return 0
}

#############################################################################
## Procedure:  read_VTS_IFO
## process the data in one VTS_xx_0.IFO of the DVD directory supplied
proc read_VTS_IFO {filename vtsnum sectorsvar pgcsvar menusvar} {

	# Open and read original IFO file in memory
	set data [readFile $filename binary]
	if {$data == ""} {return 1}

	# Verify file is of correct type
	if {[string range $data 0 11] != "DVDVIDEO-VTS"} {
		errmsg 1000 "checking IFO file type" \
				"File \"$filename\"\nis not of correct type DVDVIDEO-VTS"
		return 1000
	}

	##### split the VTS_xx_0.IFO file data in main tables parts
	set start_address(0) 0
	set end_address(0) [get_int data 0x80 4]
	for {set i 1} {$i < 9} {incr i} {
		set start_address($i) [expr [get_int data [expr 0xC4 + (4 * $i)] 4] * 0x800]
		if {$start_address($i) == 0} {
			set end_address($i) 0
		} else {
			# VTSM_VOBU_ADMAP and VTS_VOBU_ADMAP have the end_byte at position 0, not 4!
			switch $i {
				6 -
				8 {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set end_address($i) [expr $start_address($i) + \
					[get_int data [expr $start_address($i)+$end_byte_pos] 4]]
		}
	}

	set i 0
	foreach p {VTSI_MAT VTS_PTT_SRPT VTS_PGCI VTSM_PGCI_UT VTS_TMAPTI \
			VTSM_C_ADT VTSM_VOBU_ADMAP VTS_C_ADT VTS_VOBU_ADMAP} {
		if {$end_address($i) > 0} {
			set ${sectorsvar}($vtsnum,$p) [string range $data $start_address($i) \
						$end_address($i)]
		} else {
			set ${sectorsvar}($vtsnum,$p) ""
		}
		incr i
	}
	unset data

	parse_VTS_IFO $vtsnum $sectorsvar $pgcsvar $menusvar
	return 0
}

#############################################################################
## Procedure:  parse_VTS_IFO
## process the data in one VTS_xx_0.IFO
proc parse_VTS_IFO {vtsnum sectorsvar pgcsvar menusvar} {

	##### VTS_PGCI
	#dprint 2 "\nVTS_PGCI"
	parse_VTS_PGCI ${sectorsvar}($vtsnum,VTS_PGCI) $vtsnum $pgcsvar

	##### VTSM_PGCI_UT
	#dprint 2 "\nVTSM_PGCI_UT"
	parse_PGCI_UT ${sectorsvar}($vtsnum,VTSM_PGCI_UT) $vtsnum $menusvar

	return 0
}

#############################################################################
## Procedure:  readdvd
## read the DVD in memory, and refresh the interface
## args: directoryname:  Path to the VIDEO_TS directory.
##       sectorsvar, pgcsvar, menusvar: name of variables to store the IFOs in.
## Variables usage:
##
## pgcs(numvts)								;# number of VTS in the DVD
## pgcs($vtsnum,numpgcs)					;# number of PGCs in this VTS
## pgcs($vtsnum,$pgcnum)					;# list of commands in this PGC
## pgcs($vtsnum,$pgcnum,Title)				;# Title number for this PGC
## pgcs($vtsnum,$pgcnum,header)				;# Header part of the PGC
## pgcs($vtsnum,$pgcnum,numprograms)		;# Number of Programs in the PGC
## pgcs($vtsnum,$pgcnum,programmap)			;# PGC program map, in binary
## pgcs($vtsnum,$pgcnum,numcells)			;# Number of Cells in the PGC
## pgcs($vtsnum,$pgcnum,cellpbtable)		;# Cell Playback Table, in binary
## pgcs($vtsnum,$pgcnum,cellpostable)		;# Cell Position Table, in binary
##
## menus(numvts)							;# number of VTS in the DVD
## menus($vtsnum,numlus)					;# number of LUs in this VTS (0=no menu)
## menus($vtsnum,$lunum,lang)				;# language code for this LU
## menus($vtsnum,$lunum,numpgcs)			;# number of PGCs in this LU
## menus($vtsnum,$lunum,$pgcnum,mask)		;# category mask value for this PGC
## menus($vtsnum,$lunum,$pgcnum,masktext)	;# mask in human readeable form, or ""
## menus($vtsnum,$lunum,$pgcnum)			;# list of commands in this PGC
## menus($vtsnum,$lunum,$pgcnum,header)		;# Header part of the PGC
## menus($vtsnum,$lunum,$pgcnum,numprograms);# Number of Programs in the PGC
## menus($vtsnum,$lunum,$pgcnum,programmap)	;# PGC program map, in binary
## menus($vtsnum,$lunum,$pgcnum,numcells)   ;# Number of Cells in the PGC
## menus($vtsnum,$lunum,$pgcnum,cellpbtable);# Cell Playback Table, in binary
## menus($vtsnum,$lunum,$pgcnum,cellpostable);# Cell Position Table, in binary
##
## sectors($vtsnum,<tablename>)				;# raw, unmodified IFO data, splitted
##											;# by tables (like "VMGM_MAT")
##
## dvd_modified								;# flag: the DVD has been modified.
##
## Notes:
## - Tcl arrays are one dimention only.  The indexes are __string__ values.
##   Therefore, you cannot use spaces around the comas in the index names!
## - the array indexes that are not prefixed with $ are constant text strings.
## - The vtsnum, pgcnum, lunum and commands indexes are numbered starting at 1
## - vtsnum 0 is the VMG (VIDEO_TS.IFO)
## - the lists of commands are Tcl lists with the commands in hex form,
##   with blank between bytes: "00 00 00 00 00 00 00 00".
##   Elements beginning with * are pre, post and cell separators
##
proc readdvd {directoryname sectorsvar pgcsvar menusvar} {

	#set oldcursor [. cget -cursor]
	#. configure -cursor watch
	#update idletasks

	# unset old global variables
	set ${sectorsvar}(x) ""
	set ${pgcsvar}(x) ""
	set ${menusvar}(x) ""
	unset $sectorsvar $pgcsvar $menusvar

	set ::dvd_modified false

	# Open and read VMG IFO file
	set errcode [read_VMG_IFO [file join $directoryname "VIDEO_TS.IFO"] \
			$sectorsvar $pgcsvar $menusvar]
	if {$errcode == 0} {
		set numvts [subst $${pgcsvar}(numvts)]

		# compare stored number of VTS and real number of VTS_XX_0.IFO files.
		for {set vtsnum2 1} {$vtsnum2 <= 99} {incr vtsnum2} {
			if {! [file exists [file join $directoryname "VTS_[format %02i $vtsnum2]_0.IFO"]]} {
				incr vtsnum2 -1
				break
			}
		}
		if {$vtsnum2 != $numvts} {
			set txt "ERROR: The number of VTS stored in VIDEO_TS.IFO ($numvts)"
			set txt "$txt\ndon't match the real number of Title Sets ($vtsnum2)."
			set txt "$txt\n\nDo you want to fix the number of VTS to $vtsnum2 ?"
			set txt "$txt\nWARNING: The VMG_PTT_SRPT table will NOT be updated!"
			set r [tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
					-type yesno -default no -message $txt]
			if {$r == "yes"} {
				::writedvd::set_int ${sectorsvar}(0,VMGM_MAT) 0x3E $vtsnum2 2
				set numvts $vtsnum2
				set ${pgcsvar}(numvts)  $numvts
				set ${menusvar}(numvts) $numvts
				set ::dvd_modified true
			}
		}

		# read all VTS
		for {set vtsnum 1} {$vtsnum <= $numvts} {incr vtsnum} {
			set errcode [read_VTS_IFO \
					[file join $directoryname "VTS_[format %02i $vtsnum]_0.IFO"] \
					$vtsnum $sectorsvar $pgcsvar $menusvar]
			# missing VTS
			if {$errcode != 0} {
				set txt "ERROR: Could not open the VTS $vtsnum IFO file."
				set txt "$txt\n\nDo you want to fix the number of VTS to [expr $vtsnum-1] ?"
				set txt "$txt\nWARNING: doing so will forget the Titles and Menus of the VTS $vtsnum"
				if {$vtsnum < $numvts} {set txt "$txt to $numvts"}
				set txt "${txt}."
				set txt "$txt\nThe VMG_PTT_SRPT table will be updated when you save the DVD."
				set txt "$txt\n\nIf you answer \"no\", you will need to restore the backup."
				set r [tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
						-type yesno -default no -message $txt]
				if {$r == "yes"} {
					set numvts [expr $vtsnum-1]
					::writedvd::set_int ${sectorsvar}(0,VMGM_MAT) 0x3E $numvts 2
					set ${pgcsvar}(numvts)  $numvts
					set ${menusvar}(numvts) $numvts
					set errcode 0
					set ::dvd_modified true
				}
				break
			}
		}
	}

	if {$errcode != 0} {
		set ${sectorsvar}(x) ""
		set ${pgcsvar}(x) ""
		set ${menusvar}(x) ""
		unset $sectorsvar $pgcsvar $menusvar
	}

	#. configure -cursor $oldcursor
	return $errcode
}


## read a specific IFO file.
## return 0 if file is a VMG VIDEO_TS.IFO
## return the number of the VTS if file is a VTS_XX_0.IFO file
## return -1 upon error
proc readdvdfile {filename sectorsvar pgcsvar menusvar} {

	set errnum [catch {set in [open $filename r]} err]
	if {$errnum} {
		errmsg $errnum "opening file for read" $err
		return -1
	}
	fconfigure $in -translation binary
	set errnum [catch {set data [read $in 12]} err]
	if {$errnum} {
		close $in
		errmsg $errnum "reading file" $err
		return -1
	}
	close $in

	switch $data {
		"DVDVIDEO-VMG" {
			if {[read_VMG_IFO $filename $sectorsvar $pgcsvar $menusvar] != 0} {
				return -1
			} else {
				return 0
			}
		}
		"DVDVIDEO-VTS" {
			if {[catch {set vtsnum [expr [string range [file tail $filename] 4 5] * 1]}] != 0} {
				errmsg 1000 "getting VTS number" "Invalid filename"
				return -1
			}
			if {[read_VTS_IFO $filename $vtsnum $sectorsvar $pgcsvar $menusvar] != 0} {
				return -1
			} else {
				return $vtsnum
			}
		}
		default {
			errmsg 1000 "checking IFO file type" \
					"File \"$filename\"\nis not of correct type DVDVIDEO-VMG or DVDVIDEO-VTS"
			return -1
		}
	}
}

} ;# end namespace ::readdvd
